@@ -11,10 +11,12 @@ import (
 	"strings"
 	"text/tabwriter"
 	"unicode"
+
+	"mvdan.cc/sh/v3/fileutil"
 )
 
 // PrinterOption is a function which can be passed to NewPrinter
-// to alter its behaviour. To apply option to existing Printer
+// to alter its behavior. To apply option to existing Printer
 // call it directly, for example KeepPadding(true)(printer).
 type PrinterOption func(*Printer)
 
@@ -53,6 +55,11 @@ func SpaceRedirects(enabled bool) PrinterOption {
 // Note that this feature is best-effort and will only keep the
 // alignment stable, so it may need some human help the first time it is
 // run.
+//
+// Deprecated: this formatting option is flawed and buggy, and often does
+// not result in what the user wants when the code gets complex enough.
+// The next major version, v4, will remove this feature entirely.
+// See: https://github.com/mvdan/sh/issues/658
 func KeepPadding(enabled bool) PrinterOption {
 	return func(p *Printer) {
 		if enabled && !p.keepPadding {
@@ -82,7 +89,7 @@ func Minify(enabled bool) PrinterOption {
 // newlines must still appear, such as those following comments or around
 // here-documents.
 //
-// Print's trailing newline when given a *File is not affected by this option.
+// Print's trailing newline when given a [*File] is not affected by this option.
 func SingleLine(enabled bool) PrinterOption {
 	return func(p *Printer) { p.singleLine = enabled }
 }
@@ -107,9 +114,9 @@ func NewPrinter(opts ...PrinterOption) *Printer {
 // Print "pretty-prints" the given syntax tree node to the given writer. Writes
 // to w are buffered.
 //
-// The node types supported at the moment are *File, *Stmt, *Word, any Command
-// node, and any WordPart node. A trailing newline will only be printed when a
-// *File is used.
+// The node types supported at the moment are [*File], [*Stmt], [*Word], [*Assign], any
+// [Command] node, and any WordPart node. A trailing newline will only be printed
+// when a [*File] is used.
 func (p *Printer) Print(w io.Writer, node Node) error {
 	p.reset()
 
@@ -132,22 +139,25 @@ func (p *Printer) Print(w io.Writer, node Node) error {
 	w = p.tabWriter
 
 	p.bufWriter.Reset(w)
-	switch x := node.(type) {
+	switch node := node.(type) {
 	case *File:
-		p.stmtList(x.Stmts, x.Last)
+		p.stmtList(node.Stmts, node.Last)
 		p.newline(Pos{})
 	case *Stmt:
-		p.stmtList([]*Stmt{x}, nil)
+		p.stmtList([]*Stmt{node}, nil)
 	case Command:
-		p.command(x, nil)
+		p.command(node, nil)
 	case *Word:
-		p.line = x.Pos().Line()
-		p.word(x)
+		p.line = node.Pos().Line()
+		p.word(node)
 	case WordPart:
-		p.line = x.Pos().Line()
-		p.wordPart(x, nil)
+		p.line = node.Pos().Line()
+		p.wordPart(node, nil)
+	case *Assign:
+		p.line = node.Pos().Line()
+		p.assigns([]*Assign{node})
 	default:
-		return fmt.Errorf("unsupported node type: %T", x)
+		return fmt.Errorf("unsupported node type: %T", node)
 	}
 	p.flushHeredocs()
 	p.flushComments()
@@ -211,7 +221,7 @@ func (c *colCounter) Reset(w io.Writer) {
 // Printer holds the internal state of the printing mechanism of a
 // program.
 type Printer struct {
-	bufWriter
+	bufWriter // TODO: embedding this makes the methods part of the API, which we did not intend
 	tabWriter *tabwriter.Writer
 	cols      colCounter
 
@@ -224,7 +234,8 @@ type Printer struct {
 	singleLine     bool
 	funcNextLine   bool
 
-	wantSpace   bool // space is wanted or required
+	wantSpace wantSpaceState // whether space is required or has been written
+
 	wantNewline bool // newline is wanted for pretty-printing; ignored by singleLine; ignored by singleLine
 	mustNewline bool // newline is required to keep shell syntax valid
 	wroteSemi   bool // wrote ';' for the current statement
@@ -259,7 +270,8 @@ type Printer struct {
 }
 
 func (p *Printer) reset() {
-	p.wantSpace, p.wantNewline, p.mustNewline = false, false, false
+	p.wantSpace = spaceWritten
+	p.wantNewline, p.mustNewline = false, false
 	p.pendingComments = p.pendingComments[:0]
 
 	// minification uses its own newline logic
@@ -280,7 +292,7 @@ func (p *Printer) spaces(n uint) {
 
 func (p *Printer) space() {
 	p.WriteByte(' ')
-	p.wantSpace = false
+	p.wantSpace = spaceWritten
 }
 
 func (p *Printer) spacePad(pos Pos) {
@@ -289,9 +301,9 @@ func (p *Printer) spacePad(pos Pos) {
 		// with spaces, since this may result in mixing of spaces and tabs.
 		return
 	}
-	if p.wantSpace {
+	if p.wantSpace == spaceRequired {
 		p.WriteByte(' ')
-		p.wantSpace = false
+		p.wantSpace = spaceWritten
 	}
 	for p.cols.column > 0 && p.cols.column < int(pos.Col()) {
 		p.WriteByte(' ')
@@ -301,22 +313,27 @@ func (p *Printer) spacePad(pos Pos) {
 // wantsNewline reports whether we want to print at least one newline before
 // printing a node at a given position. A zero position can be given to simply
 // tell if we want a newline following what's just been printed.
-func (p *Printer) wantsNewline(pos Pos) bool {
+func (p *Printer) wantsNewline(pos Pos, escapingNewline bool) bool {
 	if p.mustNewline {
 		// We must have a newline here.
 		return true
 	}
-	if p.singleLine {
-		// The newline is optional, and singleLine turns it off.
+	if p.singleLine && len(p.pendingComments) == 0 {
+		// The newline is optional, and singleLine skips it.
+		// Don't skip if there are any pending comments,
+		// as that might move them further down to the wrong place.
 		return false
 	}
-	// THe newline is optional, and we want it via either wantNewline or via
+	if escapingNewline && p.minify {
+		return false
+	}
+	// The newline is optional, and we want it via either wantNewline or via
 	// the position's line.
 	return p.wantNewline || pos.Line() > p.line
 }
 
 func (p *Printer) bslashNewl() {
-	if p.wantSpace {
+	if p.wantSpace == spaceRequired {
 		p.space()
 	}
 	p.WriteString("\\\n")
@@ -327,22 +344,22 @@ func (p *Printer) bslashNewl() {
 func (p *Printer) spacedString(s string, pos Pos) {
 	p.spacePad(pos)
 	p.WriteString(s)
-	p.wantSpace = true
+	p.wantSpace = spaceRequired
 }
 
 func (p *Printer) spacedToken(s string, pos Pos) {
 	if p.minify {
 		p.WriteString(s)
-		p.wantSpace = false
+		p.wantSpace = spaceNotRequired
 		return
 	}
 	p.spacePad(pos)
 	p.WriteString(s)
-	p.wantSpace = true
+	p.wantSpace = spaceRequired
 }
 
 func (p *Printer) semiOrNewl(s string, pos Pos) {
-	if p.wantsNewline(Pos{}) {
+	if p.wantsNewline(Pos{}, false) {
 		p.newline(pos)
 		p.indent()
 	} else {
@@ -352,10 +369,10 @@ func (p *Printer) semiOrNewl(s string, pos Pos) {
 		if !p.minify {
 			p.space()
 		}
-		p.line = pos.Line()
+		p.advanceLine(pos.Line())
 	}
 	p.WriteString(s)
-	p.wantSpace = true
+	p.wantSpace = spaceRequired
 }
 
 func (p *Printer) writeLit(s string) {
@@ -408,13 +425,19 @@ func (p *Printer) indent() {
 
 // TODO(mvdan): add an indent call at the end of newline?
 
+// newline prints one newline and advances p.line to pos.Line().
 func (p *Printer) newline(pos Pos) {
 	p.flushHeredocs()
 	p.flushComments()
 	p.WriteByte('\n')
-	p.wantNewline, p.mustNewline, p.wantSpace = false, false, false
-	if p.line < pos.Line() {
-		p.line++
+	p.wantSpace = spaceWritten
+	p.wantNewline, p.mustNewline = false, false
+	p.advanceLine(pos.Line())
+}
+
+func (p *Printer) advanceLine(line uint) {
+	if p.line < line {
+		p.line = line
 	}
 }
 
@@ -445,7 +468,8 @@ func (p *Printer) flushHeredocs() {
 	for _, r := range hdocs {
 		p.line++
 		p.WriteByte('\n')
-		p.wantNewline, p.wantNewline, p.wantSpace = false, false, false
+		p.wantSpace = spaceWritten
+		p.wantNewline, p.wantNewline = false, false
 		if r.Op == DashHdoc && p.indentSpaces == 0 && !p.minify {
 			if r.Hdoc != nil {
 				extra := extraIndenter{
@@ -477,44 +501,49 @@ func (p *Printer) flushHeredocs() {
 		if r.Hdoc != nil {
 			// Overwrite p.line, since printing r.Word again can set
 			// p.line to the beginning of the heredoc again.
-			p.line = r.Hdoc.End().Line()
+			p.advanceLine(r.Hdoc.End().Line())
 		}
-		p.wantSpace = false
+		p.wantSpace = spaceNotRequired
 	}
 	p.level = newLevel
 	p.pendingComments = coms
 	p.mustNewline = true
 }
 
+// newline prints between zero and two newlines.
+// If any newlines are printed, it advances p.line to pos.Line().
 func (p *Printer) newlines(pos Pos) {
 	if p.firstLine && len(p.pendingComments) == 0 {
 		p.firstLine = false
 		return // no empty lines at the top
 	}
-	if !p.wantsNewline(pos) {
+	if !p.wantsNewline(pos, false) {
 		return
 	}
-	p.newline(pos)
-	if pos.Line() > p.line {
-		if !p.minify {
-			// preserve single empty lines
-			p.WriteByte('\n')
-		}
-		p.line++
+	p.flushHeredocs()
+	p.flushComments()
+	p.WriteByte('\n')
+	p.wantSpace = spaceWritten
+	p.wantNewline, p.mustNewline = false, false
+
+	l := pos.Line()
+	if l > p.line+1 && !p.minify {
+		p.WriteByte('\n') // preserve single empty lines
 	}
+	p.advanceLine(l)
 	p.indent()
 }
 
 func (p *Printer) rightParen(pos Pos) {
-	if !p.minify {
+	if len(p.pendingHdocs) > 0 || !p.minify {
 		p.newlines(pos)
 	}
 	p.WriteByte(')')
-	p.wantSpace = true
+	p.wantSpace = spaceRequired
 }
 
 func (p *Printer) semiRsrv(s string, pos Pos) {
-	if p.wantsNewline(pos) {
+	if p.wantsNewline(pos, false) {
 		p.newlines(pos)
 	} else {
 		if !p.wroteSemi {
@@ -525,7 +554,7 @@ func (p *Printer) semiRsrv(s string, pos Pos) {
 		}
 	}
 	p.WriteString(s)
-	p.wantSpace = true
+	p.wantSpace = spaceRequired
 }
 
 func (p *Printer) flushComments() {
@@ -546,19 +575,19 @@ func (p *Printer) flushComments() {
 				p.WriteByte('\n')
 			}
 			p.indent()
-			p.wantSpace = false
+			p.wantSpace = spaceWritten
 			p.spacePad(c.Pos())
-		case p.wantSpace:
+		case p.wantSpace == spaceRequired:
 			if p.keepPadding {
 				p.spacePad(c.Pos())
 			} else {
 				p.WriteByte('\t')
 			}
+		case p.wantSpace != spaceWritten:
+			p.space()
 		}
 		// don't go back one line, which may happen in some edge cases
-		if p.line < cline {
-			p.line = cline
-		}
+		p.advanceLine(cline)
 		p.WriteByte('#')
 		p.writeLit(strings.TrimRightFunc(c.Text, unicode.IsSpace))
 		p.wantNewline = true
@@ -569,106 +598,130 @@ func (p *Printer) flushComments() {
 
 func (p *Printer) comments(comments ...Comment) {
 	if p.minify {
+		for _, c := range comments {
+			if fileutil.Shebang([]byte("#"+c.Text)) != "" && c.Hash.Col() == 1 && c.Hash.Line() == 1 {
+				p.WriteString(strings.TrimRightFunc("#"+c.Text, unicode.IsSpace))
+				p.WriteString("\n")
+				p.line++
+			}
+		}
 		return
 	}
 	p.pendingComments = append(p.pendingComments, comments...)
 }
 
 func (p *Printer) wordParts(wps []WordPart, quoted bool) {
+	// We disallow unquoted escaped newlines between word parts below.
+	// However, we want to allow a leading escaped newline for cases such as:
+	//
+	//   foo <<< \
+	//     "bar baz"
+	if !quoted && !p.singleLine && wps[0].Pos().Line() > p.line {
+		p.bslashNewl()
+	}
 	for i, wp := range wps {
 		var next WordPart
 		if i+1 < len(wps) {
 			next = wps[i+1]
 		}
+		// Keep escaped newlines separating word parts when quoted.
+		// Note that those escaped newlines don't cause indentaiton.
+		// When not quoted, we strip them out consistently,
+		// because attempting to keep them would prevent indentation.
 		// Can't use p.wantsNewline here, since this is only about
 		// escaped newlines.
-		for !p.singleLine && wp.Pos().Line() > p.line {
-			if quoted {
-				// No extra spacing or indentation if quoted.
-				p.WriteString("\\\n")
-				p.line++
-			} else {
-				p.bslashNewl()
-			}
+		for quoted && !p.singleLine && wp.Pos().Line() > p.line {
+			p.WriteString("\\\n")
+			p.line++
 		}
 		p.wordPart(wp, next)
-		p.line = wp.End().Line()
+		p.advanceLine(wp.End().Line())
 	}
 }
 
 func (p *Printer) wordPart(wp, next WordPart) {
-	switch x := wp.(type) {
+	switch wp := wp.(type) {
 	case *Lit:
-		p.writeLit(x.Value)
+		p.writeLit(wp.Value)
 	case *SglQuoted:
-		if x.Dollar {
+		if wp.Dollar {
 			p.WriteByte('$')
 		}
 		p.WriteByte('\'')
-		p.writeLit(x.Value)
+		p.writeLit(wp.Value)
 		p.WriteByte('\'')
-		p.line = x.End().Line()
+		p.advanceLine(wp.End().Line())
 	case *DblQuoted:
-		p.dblQuoted(x)
+		p.dblQuoted(wp)
 	case *CmdSubst:
-		p.line = x.Pos().Line()
+		p.advanceLine(wp.Pos().Line())
 		switch {
-		case x.TempFile:
+		case wp.TempFile:
 			p.WriteString("${")
-			p.wantSpace = true
-			p.nestedStmts(x.Stmts, x.Last, x.Right)
-			p.wantSpace = false
-			p.semiRsrv("}", x.Right)
-		case x.ReplyVar:
+			p.wantSpace = spaceRequired
+			p.nestedStmts(wp.Stmts, wp.Last, wp.Right)
+			p.wantSpace = spaceNotRequired
+			p.semiRsrv("}", wp.Right)
+		case wp.ReplyVar:
 			p.WriteString("${|")
-			p.nestedStmts(x.Stmts, x.Last, x.Right)
-			p.wantSpace = false
-			p.semiRsrv("}", x.Right)
+			p.nestedStmts(wp.Stmts, wp.Last, wp.Right)
+			p.wantSpace = spaceNotRequired
+			p.semiRsrv("}", wp.Right)
+		// Special case: `# inline comment`
+		case wp.Backquotes && len(wp.Stmts) == 0 &&
+			len(wp.Last) == 1 && wp.Right.Line() == p.line:
+			p.WriteString("`#")
+			p.WriteString(wp.Last[0].Text)
+			p.WriteString("`")
 		default:
 			p.WriteString("$(")
-			p.wantSpace = len(x.Stmts) > 0 && startsWithLparen(x.Stmts[0])
-			p.nestedStmts(x.Stmts, x.Last, x.Right)
-			p.rightParen(x.Right)
+			if len(wp.Stmts) > 0 && startsWithLparen(wp.Stmts[0]) {
+				p.wantSpace = spaceRequired
+			} else {
+				p.wantSpace = spaceNotRequired
+			}
+			p.nestedStmts(wp.Stmts, wp.Last, wp.Right)
+			p.rightParen(wp.Right)
 		}
 	case *ParamExp:
 		litCont := ";"
 		if nextLit, ok := next.(*Lit); ok && nextLit.Value != "" {
 			litCont = nextLit.Value[:1]
 		}
-		name := x.Param.Value
+		name := wp.Param.Value
 		switch {
 		case !p.minify:
-		case x.Excl, x.Length, x.Width:
-		case x.Index != nil, x.Slice != nil:
-		case x.Repl != nil, x.Exp != nil:
+		case wp.Excl, wp.Length, wp.Width:
+		case wp.Index != nil, wp.Slice != nil:
+		case wp.Repl != nil, wp.Exp != nil:
 		case len(name) > 1 && !ValidName(name): // ${10}
 		case ValidName(name + litCont): // ${var}cont
 		default:
-			x2 := *x
+			x2 := *wp
 			x2.Short = true
 			p.paramExp(&x2)
 			return
 		}
-		p.paramExp(x)
+		p.paramExp(wp)
 	case *ArithmExp:
 		p.WriteString("$((")
-		if x.Unsigned {
+		if wp.Unsigned {
 			p.WriteString("# ")
 		}
-		p.arithmExpr(x.X, false, false)
+		p.arithmExpr(wp.X, false, false)
 		p.WriteString("))")
 	case *ExtGlob:
-		p.WriteString(x.Op.String())
-		p.writeLit(x.Pattern.Value)
+		p.WriteString(wp.Op.String())
+		p.writeLit(wp.Pattern.Value)
 		p.WriteByte(')')
 	case *ProcSubst:
 		// avoid conflict with << and others
-		if p.wantSpace {
+		if p.wantSpace == spaceRequired {
 			p.space()
 		}
-		p.WriteString(x.Op.String())
-		p.nestedStmts(x.Stmts, x.Last, x.Rparen)
-		p.rightParen(x.Rparen)
+		p.WriteString(wp.Op.String())
+		p.nestedStmts(wp.Stmts, wp.Last, wp.Rparen)
+		p.rightParen(wp.Rparen)
 	}
 }
 
@@ -753,23 +806,23 @@ func (p *Printer) paramExp(pe *ParamExp) {
 }
 
 func (p *Printer) loop(loop Loop) {
-	switch x := loop.(type) {
+	switch loop := loop.(type) {
 	case *WordIter:
-		p.writeLit(x.Name.Value)
-		if x.InPos.IsValid() {
+		p.writeLit(loop.Name.Value)
+		if loop.InPos.IsValid() {
 			p.spacedString(" in", Pos{})
-			p.wordJoin(x.Items)
+			p.wordJoin(loop.Items)
 		}
 	case *CStyleLoop:
 		p.WriteString("((")
-		if x.Init == nil {
+		if loop.Init == nil {
 			p.space()
 		}
-		p.arithmExpr(x.Init, false, false)
+		p.arithmExpr(loop.Init, false, false)
 		p.WriteString("; ")
-		p.arithmExpr(x.Cond, false, false)
+		p.arithmExpr(loop.Cond, false, false)
 		p.WriteString("; ")
-		p.arithmExpr(x.Post, false, false)
+		p.arithmExpr(loop.Post, false, false)
 		p.WriteString("))")
 	}
 }
@@ -778,40 +831,40 @@ func (p *Printer) arithmExpr(expr ArithmExpr, compact, spacePlusMinus bool) {
 	if p.minify {
 		compact = true
 	}
-	switch x := expr.(type) {
+	switch expr := expr.(type) {
 	case *Word:
-		p.word(x)
+		p.word(expr)
 	case *BinaryArithm:
 		if compact {
-			p.arithmExpr(x.X, compact, spacePlusMinus)
-			p.WriteString(x.Op.String())
-			p.arithmExpr(x.Y, compact, false)
+			p.arithmExpr(expr.X, compact, spacePlusMinus)
+			p.WriteString(expr.Op.String())
+			p.arithmExpr(expr.Y, compact, false)
 		} else {
-			p.arithmExpr(x.X, compact, spacePlusMinus)
-			if x.Op != Comma {
+			p.arithmExpr(expr.X, compact, spacePlusMinus)
+			if expr.Op != Comma {
 				p.space()
 			}
-			p.WriteString(x.Op.String())
+			p.WriteString(expr.Op.String())
 			p.space()
-			p.arithmExpr(x.Y, compact, false)
+			p.arithmExpr(expr.Y, compact, false)
 		}
 	case *UnaryArithm:
-		if x.Post {
-			p.arithmExpr(x.X, compact, spacePlusMinus)
-			p.WriteString(x.Op.String())
+		if expr.Post {
+			p.arithmExpr(expr.X, compact, spacePlusMinus)
+			p.WriteString(expr.Op.String())
 		} else {
 			if spacePlusMinus {
-				switch x.Op {
+				switch expr.Op {
 				case Plus, Minus:
 					p.space()
 				}
 			}
-			p.WriteString(x.Op.String())
-			p.arithmExpr(x.X, compact, false)
+			p.WriteString(expr.Op.String())
+			p.arithmExpr(expr.X, compact, false)
 		}
 	case *ParenArithm:
 		p.WriteByte('(')
-		p.arithmExpr(x.X, false, false)
+		p.arithmExpr(expr.X, false, false)
 		p.WriteByte(')')
 	}
 }
@@ -821,58 +874,62 @@ func (p *Printer) testExpr(expr TestExpr) {
 	if expr.Pos().Line() > p.line {
 		p.newlines(expr.Pos())
 		p.spacePad(expr.Pos())
-	} else if p.wantSpace {
+	} else if p.wantSpace == spaceRequired {
 		p.space()
 	}
 	p.testExprSameLine(expr)
 }
 
 func (p *Printer) testExprSameLine(expr TestExpr) {
-	p.line = expr.Pos().Line()
-	switch x := expr.(type) {
+	p.advanceLine(expr.Pos().Line())
+	switch expr := expr.(type) {
 	case *Word:
-		p.word(x)
+		p.word(expr)
 	case *BinaryTest:
-		p.testExprSameLine(x.X)
+		p.testExprSameLine(expr.X)
 		p.space()
-		p.WriteString(x.Op.String())
-		switch x.Op {
+		p.WriteString(expr.Op.String())
+		switch expr.Op {
 		case AndTest, OrTest:
-			p.wantSpace = true
-			p.testExpr(x.Y)
+			p.wantSpace = spaceRequired
+			p.testExpr(expr.Y)
 		default:
 			p.space()
-			p.testExprSameLine(x.Y)
+			p.testExprSameLine(expr.Y)
 		}
 	case *UnaryTest:
-		p.WriteString(x.Op.String())
+		p.WriteString(expr.Op.String())
 		p.space()
-		p.testExprSameLine(x.X)
+		p.testExprSameLine(expr.X)
 	case *ParenTest:
 		p.WriteByte('(')
-		p.wantSpace = startsWithLparen(x.X)
-		p.testExpr(x.X)
+		if startsWithLparen(expr.X) {
+			p.wantSpace = spaceRequired
+		} else {
+			p.wantSpace = spaceNotRequired
+		}
+		p.testExpr(expr.X)
 		p.WriteByte(')')
 	}
 }
 
 func (p *Printer) word(w *Word) {
 	p.wordParts(w.Parts, false)
-	p.wantSpace = true
+	p.wantSpace = spaceRequired
 }
 
 func (p *Printer) unquotedWord(w *Word) {
 	for _, wp := range w.Parts {
-		switch x := wp.(type) {
+		switch wp := wp.(type) {
 		case *SglQuoted:
-			p.writeLit(x.Value)
+			p.writeLit(wp.Value)
 		case *DblQuoted:
-			p.wordParts(x.Parts, true)
+			p.wordParts(wp.Parts, true)
 		case *Lit:
-			for i := 0; i < len(x.Value); i++ {
-				if b := x.Value[i]; b == '\\' {
-					if i++; i < len(x.Value) {
-						p.WriteByte(x.Value[i])
+			for i := 0; i < len(wp.Value); i++ {
+				if b := wp.Value[i]; b == '\\' {
+					if i++; i < len(wp.Value) {
+						p.WriteByte(wp.Value[i])
 					}
 				} else {
 					p.WriteByte(b)
@@ -906,7 +963,7 @@ func (p *Printer) casePatternJoin(pats []*Word) {
 		if i > 0 {
 			p.spacedToken("|", Pos{})
 		}
-		if p.wantsNewline(w.Pos()) {
+		if p.wantsNewline(w.Pos(), true) {
 			if !anyNewline {
 				p.incLevel()
 				anyNewline = true
@@ -937,7 +994,7 @@ func (p *Printer) elemJoin(elems []*ArrayElem, last []Comment) {
 		if el.Pos().Line() > p.line {
 			p.newlines(el.Pos())
 			p.spacePad(el.Pos())
-		} else if p.wantSpace {
+		} else if p.wantSpace == spaceRequired {
 			p.space()
 		}
 		if p.wroteIndex(el.Index) {
@@ -966,10 +1023,10 @@ func (p *Printer) stmt(s *Stmt) {
 	}
 	p.incLevel()
 	for _, r := range s.Redirs[startRedirs:] {
-		if p.wantsNewline(r.OpPos) {
+		if p.wantsNewline(r.OpPos, true) {
 			p.bslashNewl()
 		}
-		if p.wantSpace {
+		if p.wantSpace == spaceRequired {
 			p.spacePad(r.Pos())
 		}
 		if r.N != nil {
@@ -979,7 +1036,7 @@ func (p *Printer) stmt(s *Stmt) {
 		if p.spaceRedirects && (r.Op != DplIn && r.Op != DplOut) {
 			p.space()
 		} else {
-			p.wantSpace = true
+			p.wantSpace = spaceRequired
 		}
 		p.word(r.Word)
 		if r.Op == Hdoc || r.Op == DashHdoc {
@@ -1001,86 +1058,109 @@ func (p *Printer) stmt(s *Stmt) {
 			p.WriteString(";")
 		}
 		p.wroteSemi = true
-		p.wantSpace = true
+		p.wantSpace = spaceRequired
 	}
 	p.decLevel()
 }
 
+func (p *Printer) printRedirsUntil(redirs []*Redirect, startRedirs int, pos Pos) int {
+	for _, r := range redirs[startRedirs:] {
+		if r.Pos().After(pos) || r.Op == Hdoc || r.Op == DashHdoc {
+			break
+		}
+		if p.wantSpace == spaceRequired {
+			p.spacePad(r.Pos())
+		}
+		if r.N != nil {
+			p.writeLit(r.N.Value)
+		}
+		p.WriteString(r.Op.String())
+		if p.spaceRedirects && (r.Op != DplIn && r.Op != DplOut) {
+			p.space()
+		} else {
+			p.wantSpace = spaceRequired
+		}
+		p.word(r.Word)
+		startRedirs++
+	}
+	return startRedirs
+}
+
 func (p *Printer) command(cmd Command, redirs []*Redirect) (startRedirs int) {
-	p.line = cmd.Pos().Line()
+	p.advanceLine(cmd.Pos().Line())
 	p.spacePad(cmd.Pos())
-	switch x := cmd.(type) {
+	switch cmd := cmd.(type) {
 	case *CallExpr:
-		p.assigns(x.Assigns)
-		if len(x.Args) <= 1 {
-			p.wordJoin(x.Args)
-			return 0
+		p.assigns(cmd.Assigns)
+		if len(cmd.Args) > 0 {
+			startRedirs = p.printRedirsUntil(redirs, startRedirs, cmd.Args[0].Pos())
 		}
-		p.wordJoin(x.Args[:1])
-		for _, r := range redirs {
-			if r.Pos().After(x.Args[1].Pos()) || r.Op == Hdoc || r.Op == DashHdoc {
-				break
-			}
-			if p.wantSpace {
-				p.spacePad(r.Pos())
-			}
-			if r.N != nil {
-				p.writeLit(r.N.Value)
-			}
-			p.WriteString(r.Op.String())
-			if p.spaceRedirects && (r.Op != DplIn && r.Op != DplOut) {
-				p.space()
-			} else {
-				p.wantSpace = true
-			}
-			p.word(r.Word)
-			startRedirs++
+		if len(cmd.Args) <= 1 {
+			p.wordJoin(cmd.Args)
+			return startRedirs
 		}
-		p.wordJoin(x.Args[1:])
+		p.wordJoin(cmd.Args[:1])
+		startRedirs = p.printRedirsUntil(redirs, startRedirs, cmd.Args[1].Pos())
+		p.wordJoin(cmd.Args[1:])
 	case *Block:
 		p.WriteByte('{')
-		p.wantSpace = true
+		p.wantSpace = spaceRequired
 		// Forbid "foo()\n{ bar; }"
 		p.wantNewline = p.wantNewline || p.funcNextLine
-		p.nestedStmts(x.Stmts, x.Last, x.Rbrace)
-		p.semiRsrv("}", x.Rbrace)
+		p.nestedStmts(cmd.Stmts, cmd.Last, cmd.Rbrace)
+		p.semiRsrv("}", cmd.Rbrace)
 	case *IfClause:
-		p.ifClause(x, false)
+		p.ifClause(cmd, false)
 	case *Subshell:
 		p.WriteByte('(')
-		p.wantSpace = len(x.Stmts) > 0 && startsWithLparen(x.Stmts[0])
-		p.spacePad(stmtsPos(x.Stmts, x.Last))
-		p.nestedStmts(x.Stmts, x.Last, x.Rparen)
-		p.wantSpace = false
-		p.spacePad(x.Rparen)
-		p.rightParen(x.Rparen)
-	case *WhileClause:
-		if x.Until {
-			p.spacedString("until", x.Pos())
+		stmts := cmd.Stmts
+		if len(stmts) > 0 && startsWithLparen(stmts[0]) {
+			p.wantSpace = spaceRequired
+			// Add a space between nested parentheses if we're printing them in a single line,
+			// to avoid the ambiguity between `((` and `( (`.
+			if (cmd.Lparen.Line() != stmts[0].Pos().Line() || len(stmts) > 1) && !p.singleLine {
+				p.wantSpace = spaceNotRequired
+
+				if p.minify {
+					p.mustNewline = true
+				}
+			}
 		} else {
-			p.spacedString("while", x.Pos())
+			p.wantSpace = spaceNotRequired
 		}
-		p.nestedStmts(x.Cond, x.CondLast, Pos{})
-		p.semiOrNewl("do", x.DoPos)
-		p.nestedStmts(x.Do, x.DoLast, x.DonePos)
-		p.semiRsrv("done", x.DonePos)
+
+		p.spacePad(stmtsPos(cmd.Stmts, cmd.Last))
+		p.nestedStmts(cmd.Stmts, cmd.Last, cmd.Rparen)
+		p.wantSpace = spaceNotRequired
+		p.spacePad(cmd.Rparen)
+		p.rightParen(cmd.Rparen)
+	case *WhileClause:
+		if cmd.Until {
+			p.spacedString("until", cmd.Pos())
+		} else {
+			p.spacedString("while", cmd.Pos())
+		}
+		p.nestedStmts(cmd.Cond, cmd.CondLast, Pos{})
+		p.semiOrNewl("do", cmd.DoPos)
+		p.nestedStmts(cmd.Do, cmd.DoLast, cmd.DonePos)
+		p.semiRsrv("done", cmd.DonePos)
 	case *ForClause:
-		if x.Select {
+		if cmd.Select {
 			p.WriteString("select ")
 		} else {
 			p.WriteString("for ")
 		}
-		p.loop(x.Loop)
-		p.semiOrNewl("do", x.DoPos)
-		p.nestedStmts(x.Do, x.DoLast, x.DonePos)
-		p.semiRsrv("done", x.DonePos)
+		p.loop(cmd.Loop)
+		p.semiOrNewl("do", cmd.DoPos)
+		p.nestedStmts(cmd.Do, cmd.DoLast, cmd.DonePos)
+		p.semiRsrv("done", cmd.DonePos)
 	case *BinaryCmd:
-		p.stmt(x.X)
-		if p.minify || p.singleLine || x.Y.Pos().Line() <= p.line {
+		p.stmt(cmd.X)
+		if p.minify || p.singleLine || cmd.Y.Pos().Line() <= p.line {
 			// leave p.nestedBinary untouched
-			p.spacedToken(x.Op.String(), x.OpPos)
-			p.line = x.Y.Pos().Line()
-			p.stmt(x.Y)
+			p.spacedToken(cmd.Op.String(), cmd.OpPos)
+			p.advanceLine(cmd.Y.Pos().Line())
+			p.stmt(cmd.Y)
 			break
 		}
 		indent := !p.nestedBinary
@@ -1091,59 +1171,60 @@ func (p *Printer) command(cmd Command, redirs []*Redirect) (startRedirs int) {
 			if len(p.pendingHdocs) == 0 {
 				p.bslashNewl()
 			}
-			p.spacedToken(x.Op.String(), x.OpPos)
-			if len(x.Y.Comments) > 0 {
-				p.wantSpace = false
-				p.newline(x.Y.Pos())
+			p.spacedToken(cmd.Op.String(), cmd.OpPos)
+			if len(cmd.Y.Comments) > 0 {
+				p.wantSpace = spaceNotRequired
+				p.newline(cmd.Y.Pos())
 				p.indent()
-				p.comments(x.Y.Comments...)
+				p.comments(cmd.Y.Comments...)
 				p.newline(Pos{})
 				p.indent()
 			}
 		} else {
-			p.spacedToken(x.Op.String(), x.OpPos)
-			p.line = x.OpPos.Line()
-			p.comments(x.Y.Comments...)
+			p.spacedToken(cmd.Op.String(), cmd.OpPos)
+			p.advanceLine(cmd.OpPos.Line())
+			p.comments(cmd.Y.Comments...)
 			p.newline(Pos{})
 			p.indent()
 		}
-		p.line = x.Y.Pos().Line()
-		_, p.nestedBinary = x.Y.Cmd.(*BinaryCmd)
-		p.stmt(x.Y)
+		p.advanceLine(cmd.Y.Pos().Line())
+		_, p.nestedBinary = cmd.Y.Cmd.(*BinaryCmd)
+		p.stmt(cmd.Y)
 		if indent {
 			p.decLevel()
 		}
 		p.nestedBinary = false
 	case *FuncDecl:
-		if x.RsrvWord {
+		if cmd.RsrvWord {
 			p.WriteString("function ")
 		}
-		p.writeLit(x.Name.Value)
-		if !x.RsrvWord || x.Parens {
+		p.writeLit(cmd.Name.Value)
+		if !cmd.RsrvWord || cmd.Parens {
 			p.WriteString("()")
 		}
 		if p.funcNextLine {
 			p.newline(Pos{})
 			p.indent()
-		} else if !x.Parens || !p.minify {
+		} else if !cmd.Parens || !p.minify {
 			p.space()
 		}
-		p.line = x.Body.Pos().Line()
-		p.comments(x.Body.Comments...)
-		p.stmt(x.Body)
+		p.advanceLine(cmd.Body.Pos().Line())
+		p.comments(cmd.Body.Comments...)
+		p.stmt(cmd.Body)
 	case *CaseClause:
 		p.WriteString("case ")
-		p.word(x.Word)
+		p.word(cmd.Word)
 		p.WriteString(" in")
-		p.wantSpace = true
+		p.advanceLine(cmd.In.Line())
+		p.wantSpace = spaceRequired
 		if p.swtCaseIndent {
 			p.incLevel()
 		}
-		if len(x.Items) == 0 {
+		if len(cmd.Items) == 0 {
 			// Apparently "case x in; esac" is invalid shell.
 			p.mustNewline = true
 		}
-		for i, ci := range x.Items {
+		for i, ci := range cmd.Items {
 			var last []Comment
 			for i, c := range ci.Comments {
 				if c.Pos().After(ci.Pos()) {
@@ -1156,7 +1237,11 @@ func (p *Printer) command(cmd Command, redirs []*Redirect) (startRedirs int) {
 			p.spacePad(ci.Pos())
 			p.casePatternJoin(ci.Patterns)
 			p.WriteByte(')')
-			p.wantSpace = !p.minify
+			if !p.minify {
+				p.wantSpace = spaceRequired
+			} else {
+				p.wantSpace = spaceNotRequired
+			}
 
 			bodyPos := stmtsPos(ci.Stmts, ci.Last)
 			bodyEnd := stmtsEnd(ci.Stmts, ci.Last)
@@ -1164,12 +1249,13 @@ func (p *Printer) command(cmd Command, redirs []*Redirect) (startRedirs int) {
 				(bodyEnd.IsValid() && ci.OpPos.Line() > bodyEnd.Line())
 			p.nestedStmts(ci.Stmts, ci.Last, ci.OpPos)
 			p.level++
-			if !p.minify || i != len(x.Items)-1 {
+			if !p.minify || i != len(cmd.Items)-1 {
 				if sep {
 					p.newlines(ci.OpPos)
 					p.wantNewline = true
 				}
 				p.spacedToken(ci.Op.String(), ci.OpPos)
+				p.advanceLine(ci.OpPos.Line())
 				// avoid ; directly after tokens like ;;
 				p.wroteSemi = true
 			}
@@ -1177,58 +1263,58 @@ func (p *Printer) command(cmd Command, redirs []*Redirect) (startRedirs int) {
 			p.flushComments()
 			p.level--
 		}
-		p.comments(x.Last...)
+		p.comments(cmd.Last...)
 		if p.swtCaseIndent {
 			p.flushComments()
 			p.decLevel()
 		}
-		p.semiRsrv("esac", x.Esac)
+		p.semiRsrv("esac", cmd.Esac)
 	case *ArithmCmd:
 		p.WriteString("((")
-		if x.Unsigned {
+		if cmd.Unsigned {
 			p.WriteString("# ")
 		}
-		p.arithmExpr(x.X, false, false)
+		p.arithmExpr(cmd.X, false, false)
 		p.WriteString("))")
 	case *TestClause:
 		p.WriteString("[[ ")
 		p.incLevel()
-		p.testExpr(x.X)
+		p.testExpr(cmd.X)
 		p.decLevel()
-		p.spacedString("]]", x.Right)
+		p.spacedString("]]", cmd.Right)
 	case *DeclClause:
-		p.spacedString(x.Variant.Value, x.Pos())
-		p.assigns(x.Args)
+		p.spacedString(cmd.Variant.Value, cmd.Pos())
+		p.assigns(cmd.Args)
 	case *TimeClause:
-		p.spacedString("time", x.Pos())
-		if x.PosixFormat {
-			p.spacedString("-p", x.Pos())
+		p.spacedString("time", cmd.Pos())
+		if cmd.PosixFormat {
+			p.spacedString("-p", cmd.Pos())
 		}
-		if x.Stmt != nil {
-			p.stmt(x.Stmt)
+		if cmd.Stmt != nil {
+			p.stmt(cmd.Stmt)
 		}
 	case *CoprocClause:
-		p.spacedString("coproc", x.Pos())
-		if x.Name != nil {
+		p.spacedString("coproc", cmd.Pos())
+		if cmd.Name != nil {
 			p.space()
-			p.word(x.Name)
+			p.word(cmd.Name)
 		}
 		p.space()
-		p.stmt(x.Stmt)
+		p.stmt(cmd.Stmt)
 	case *LetClause:
-		p.spacedString("let", x.Pos())
-		for _, n := range x.Exprs {
+		p.spacedString("let", cmd.Pos())
+		for _, n := range cmd.Exprs {
 			p.space()
 			p.arithmExpr(n, true, false)
 		}
 	case *TestDecl:
-		p.spacedString("@test", x.Pos())
+		p.spacedString("@test", cmd.Pos())
 		p.space()
-		p.word(x.Description)
+		p.word(cmd.Description)
 		p.space()
-		p.stmt(x.Body)
+		p.stmt(cmd.Body)
 	default:
-		panic(fmt.Sprintf("syntax.Printer: unexpected node type %T", x))
+		panic(fmt.Sprintf("syntax.Printer: unexpected node type %T", cmd))
 	}
 	return startRedirs
 }
@@ -1271,20 +1357,6 @@ func (p *Printer) ifClause(ic *IfClause, elif bool) {
 	p.semiRsrv("fi", ic.FiPos)
 }
 
-func startsWithLparen(node Node) bool {
-	switch node := node.(type) {
-	case *Stmt:
-		return startsWithLparen(node.Cmd)
-	case *BinaryCmd:
-		return startsWithLparen(node.X)
-	case *Subshell:
-		return true // keep ( (
-	case *ArithmCmd:
-		return true // keep ( ((
-	}
-	return false
-}
-
 func (p *Printer) stmtList(stmts []*Stmt, last []Comment) {
 	sep := p.wantNewline || (len(stmts) > 0 && stmts[0].Pos().Line() > p.line)
 	for i, s := range stmts {
@@ -1292,7 +1364,7 @@ func (p *Printer) stmtList(stmts []*Stmt, last []Comment) {
 			// In singleLine mode, ensure we use semicolons between
 			// statements.
 			p.WriteByte(';')
-			p.wantSpace = true
+			p.wantSpace = spaceRequired
 		}
 		pos := s.Pos()
 		var midComs, endComs []Comment
@@ -1313,10 +1385,10 @@ func (p *Printer) stmtList(stmts []*Stmt, last []Comment) {
 			// statement.
 			p.comments(c)
 		}
-		if !p.minify || p.wantSpace {
+		if p.mustNewline || !p.minify || p.wantSpace == spaceRequired {
 			p.newlines(pos)
 		}
-		p.line = pos.Line()
+		p.advanceLine(pos.Line())
 		p.comments(midComs...)
 		p.stmt(s)
 		p.comments(endComs...)
@@ -1328,6 +1400,75 @@ func (p *Printer) stmtList(stmts []*Stmt, last []Comment) {
 	p.comments(last...)
 }
 
+func (p *Printer) nestedStmts(stmts []*Stmt, last []Comment, closing Pos) {
+	p.incLevel()
+	switch {
+	case len(stmts) > 1:
+		// Force a newline if we find:
+		//     { stmt; stmt; }
+		p.wantNewline = true
+	case closing.Line() > p.line && len(stmts) > 0 &&
+		stmtsEnd(stmts, last).Line() < closing.Line():
+		// Force a newline if we find:
+		//     { stmt
+		//     }
+		p.wantNewline = true
+	case len(p.pendingComments) > 0 && len(stmts) > 0:
+		// Force a newline if we find:
+		//     for i in a b # stmt
+		//     do foo; done
+		p.wantNewline = true
+	}
+	p.stmtList(stmts, last)
+	if closing.IsValid() {
+		p.flushComments()
+	}
+	p.decLevel()
+}
+
+func (p *Printer) assigns(assigns []*Assign) {
+	p.incLevel()
+	for _, a := range assigns {
+		if p.wantsNewline(a.Pos(), true) {
+			p.bslashNewl()
+		} else {
+			p.spacePad(a.Pos())
+		}
+		if a.Name != nil {
+			p.writeLit(a.Name.Value)
+			p.wroteIndex(a.Index)
+			if a.Append {
+				p.WriteByte('+')
+			}
+			if !a.Naked {
+				p.WriteByte('=')
+			}
+		}
+		if a.Value != nil {
+			// Ensure we don't use an escaped newline after '=',
+			// because that can result in indentation, thus
+			// splitting "foo=bar" into "foo= bar".
+			p.advanceLine(a.Value.Pos().Line())
+			p.word(a.Value)
+		} else if a.Array != nil {
+			p.wantSpace = spaceNotRequired
+			p.WriteByte('(')
+			p.elemJoin(a.Array.Elems, a.Array.Last)
+			p.rightParen(a.Array.Rparen)
+		}
+		p.wantSpace = spaceRequired
+	}
+	p.decLevel()
+}
+
+type wantSpaceState uint8
+
+const (
+	spaceNotRequired wantSpaceState = iota
+	spaceRequired                   // we should generally print a space or a newline next
+	spaceWritten                    // we have just written a space or newline
+)
+
 // extraIndenter ensures that all lines in a '<<-' heredoc body have at least
 // baseIndent leading tabs. Those that had more tab indentation than the first
 // heredoc line will keep that relative indentation.
@@ -1371,7 +1512,7 @@ func (e *extraIndenter) WriteByte(b byte) error {
 		lineIndent += e.firstChange
 	}
 	e.bufWriter.WriteByte(tabwriter.Escape)
-	for i := 0; i < lineIndent; i++ {
+	for range lineIndent {
 		e.bufWriter.WriteByte('\t')
 	}
 	e.bufWriter.WriteByte(tabwriter.Escape)
@@ -1381,69 +1522,22 @@ func (e *extraIndenter) WriteByte(b byte) error {
 }
 
 func (e *extraIndenter) WriteString(s string) (int, error) {
-	for i := 0; i < len(s); i++ {
+	for i := range len(s) {
 		e.WriteByte(s[i])
 	}
 	return len(s), nil
 }
 
-func (p *Printer) nestedStmts(stmts []*Stmt, last []Comment, closing Pos) {
-	p.incLevel()
-	switch {
-	case len(stmts) > 1:
-		// Force a newline if we find:
-		//     { stmt; stmt; }
-		p.wantNewline = true
-	case closing.Line() > p.line && len(stmts) > 0 &&
-		stmtsEnd(stmts, last).Line() < closing.Line():
-		// Force a newline if we find:
-		//     { stmt
-		//     }
-		p.wantNewline = true
-	case len(p.pendingComments) > 0 && len(stmts) > 0:
-		// Force a newline if we find:
-		//     for i in a b # stmt
-		//     do foo; done
-		p.wantNewline = true
+func startsWithLparen(node Node) bool {
+	switch node := node.(type) {
+	case *Stmt:
+		return startsWithLparen(node.Cmd)
+	case *BinaryCmd:
+		return startsWithLparen(node.X)
+	case *Subshell:
+		return true // keep ( (
+	case *ArithmCmd:
+		return true // keep ( ((
 	}
-	p.stmtList(stmts, last)
-	if closing.IsValid() {
-		p.flushComments()
-	}
-	p.decLevel()
-}
-
-func (p *Printer) assigns(assigns []*Assign) {
-	p.incLevel()
-	for _, a := range assigns {
-		if p.wantsNewline(a.Pos()) {
-			p.bslashNewl()
-		} else {
-			p.spacePad(a.Pos())
-		}
-		if a.Name != nil {
-			p.writeLit(a.Name.Value)
-			p.wroteIndex(a.Index)
-			if a.Append {
-				p.WriteByte('+')
-			}
-			if !a.Naked {
-				p.WriteByte('=')
-			}
-		}
-		if a.Value != nil {
-			// Ensure we don't use an escaped newline after '=',
-			// because that can result in indentation, thus
-			// splitting "foo=bar" into "foo= bar".
-			p.line = a.Value.Pos().Line()
-			p.word(a.Value)
-		} else if a.Array != nil {
-			p.wantSpace = false
-			p.WriteByte('(')
-			p.elemJoin(a.Array.Elems, a.Array.Last)
-			p.rightParen(a.Array.Rparen)
-		}
-		p.wantSpace = true
-	}
-	p.decLevel()
+	return false
 }
